(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation in Standard ML.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(*require "order_sig.sml";*)
(*require "searchtree_sig.sml";*)
(*require "tree.sml";*)
(*require "$.basis.__list";*)

(* This is a general seach tree functor. It requires
   a structure with an ordered datatype as the key *)
functor FSearchTree (Key : ORDER) : SEARCH_TREE =
  struct
    exception Insert (* FIXME: Must be a general exception *)
    exception Delete (* FIXME: ... *)  
 
    type key = Key.t
    type 'a t = (key * 'a) Tree.t

    fun lookup Tree.leaf _ = NONE
      | lookup (Tree.node((foundKey, item), t1, t2)) key =
          case Key.compare (key, foundKey) of
            GREATER => (lookup t1 key)
          | EQUAL => SOME (item)
          | LESS => (lookup t2 key)

    (* General internal function used for implementing
       other functions of the structure *)
    fun mapNode f Tree.leaf _ = (f Tree.leaf)
      | mapNode f (n as Tree.node(e as (foundKey, _), t1, t2)) key =
          case Key.compare (key, foundKey) of
            GREATER => (Tree.node (e, mapNode f t1 key, t2))
          | EQUAL => (f n)
          | LESS => (Tree.node (e, t1, mapNode f t2 key))

    fun fInsert value Tree.leaf =
          Tree.node (value, Tree.leaf, Tree.leaf)
      | fInsert _ (Tree.node (_, _, _)) = raise Insert

    fun insert tree key item =
          mapNode (fInsert (key, item)) tree key

    fun fUpdate f Tree.leaf = Tree.leaf
      | fUpdate f (Tree.node ((foundKey, item), t1, t2)) =
          Tree.node ((foundKey, f item), t1, t2)

    fun update tree key f = 
          mapNode (fUpdate f) tree key

    fun fDelete (Tree.leaf) key NONE = raise Delete
      | fDelete (Tree.leaf) key (SOME(carry)) = carry
      | fDelete (Tree.node(e as (foundKey, _), t1, t2)) key carry =
          case Key.compare (key, foundKey) of
            GREATER => (Tree.node (e, fDelete t1 key carry, t2))
          | EQUAL => (fDelete t1 key (SOME(t2)))
          | LESS => (Tree.node (e, t1, fDelete t2 key carry))

    (* Delete the item associated with a given key *)
    fun delete tree key = fDelete tree key NONE

    fun map f Tree.leaf = Tree.leaf
      | map f (Tree.node (e as (k, _), t1, t2)) =
          Tree.node ((k, f e), map f t1, map f t2)

    (* Notice: This is NOT a tree fold but a linear
       fold similar to fold on a list.
       I use an inorder fold but the user may not
       rely on any particular ordering of the fold. *)
    fun fold f acc tree =
          Tree.inFold f acc tree

    (* filter reconstructs the searchtree, since this
       is easiest to implement. *)
    fun fFilter f (element, acc) =
          if f element then
            element::acc
          else
            acc

    fun fBuild ((key, value), accTree) =
          insert accTree key value 

    fun filter f tree =
          List.foldl fBuild Tree.empty (fold (fFilter f) nil tree)
  end
