(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation in Standard ML.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* FIXME: This implementation could still use some tidying up... *)

(*require "$.errorhandling.commonerrorreporters";*)
(*require "$.errorhandling.errorreporter_sig";*)
(*require "$.typeneutralwinnt.classes";*)
(*require "$.miscstringswinnt.stringres";*)
(*require "$.basis.__string";*)
(*require "$.basis.__list";*)

signature DG_NODES =
  sig
    eqtype t

    (* FIXME: Later we want to be able to remove notifiers again *)
    type notifyID = unit ref
    
    val create : string -> ClassesImpl.object -> t
    val typeName : t -> string
    val protocol : (ClassesImpl.object -> 'a) -> t -> 'a
    val protocolName : t -> string

    val isConnected : t -> t -> bool
    val connect : t -> t -> unit
    val disconnect : t -> t -> unit

    val connected : t -> t list
    val connectedToType : t -> string -> t list
    val singleConnectedToType : t -> string -> t

    val reverseConnected : t -> t list
    val reverseConnectedFromType : t -> string -> t list
    val singleReverseConnectedFromType : t -> string -> t

    (* Notifications *)
    val addConnectNotify : t -> (t -> unit) -> notifyID
    val addDisconnectNotify : t -> t -> (t -> unit) -> notifyID
  end

(* NOTICE: The error function in Err is expected to raise
   an exception in this functor *)
functor FDGNodes(Err : ERROR_REPORTER) :> DG_NODES =
  struct
    (* FIXME: Later we want to be able to remove notifiers again *)
    type notifyID = unit ref

    (* Private convenience functions for named values *)
    (* For a list-fold operation *)
    fun pRemoveNodeNotifiers theNode ((a, b), (values, notifiers)) =
          if a = theNode then (values, SOME(b))
          else ((a, b)::values, notifiers)

    fun pRemoveNode theNode (a, values) =
          if a = theNode then values
          else a::values

    (* FIXME: More general? *)
    type 'a notifiers = (('a -> unit) * notifyID) list

    (* The datatype for nodes in the hypergraph *)
    datatype t =
      node of {typeName : StringRes.t,
               protocol : ClassesImpl.object,
                 (* FIXME: Add info about how nodes may be connected *)
                 (* List of possible connections *)
                 (* FIXME: Add a dictionary for connection types *)
               connections : ((t * (t notifiers)) list ref) *
                             ((t notifiers) ref),
                 (* List of nodes being reverse connected from an attr. *)
                 (* FIXME: Add a dictionary for connection types *)
               reverseConnections : (t list) ref} ref
               (* notifyConnected *)

    (* private convenience functions for accessing data in the node type *)
    fun pNodeRec matchF (node(value)) = matchF(!value)

    val pConnections =
          pNodeRec (fn {connections, ...} => connections)

    val pReverseConnections =
          pNodeRec (fn {reverseConnections, ...} =>
                    reverseConnections)

    (* Function for reading node name *)
    fun typeName theNode =
          StringRes.value (pNodeRec (fn {typeName, ...} => typeName) theNode)

    (* Typecast to the specific protocol type for this node *)
    fun protocol typeCastF theNode =
          typeCastF (pNodeRec (fn {protocol, ...} => protocol) theNode)

    fun protocolName theNode =
          ClassesImpl.className
            (pNodeRec (fn {protocol, ...} => protocol) theNode)
          
    (* Node constructor - used directly only by node implementors *)
    fun create typeName protocol =
          node(ref {typeName = StringRes.create typeName,
                    protocol = protocol,
                    connections = (ref nil, ref nil),
                    reverseConnections = ref nil})

    (* Function for making notifications *)
    fun doNotify notifiers theNode =
          List.app (fn (notifier, _) => notifier theNode) notifiers

    (* Private convenience functions for making connections internally *)
    fun pConnect thisNode toNode =
          let
            val (nodes, notifiers) =
                   pConnections thisNode
          in
            (* FIXME: Also check if the node exists! *)
            (nodes := (toNode, nil)::(!nodes);
             doNotify (!notifiers))
          end

    fun pDisconnect thisNode toNode =
          let
            val (nodes, _) =
                   pConnections thisNode
            val (newNodes, notifiers) = foldl (pRemoveNodeNotifiers toNode) (nil, NONE) (!nodes)
            val _ = nodes := newNodes
          in
            case notifiers of
              NONE => (Err.error ("Cannot disconnect because there is no connection from " ^
                         (typeName thisNode) ^ " to " ^ (typeName toNode));
                       fn _ => ())
            | SOME(notifiers) => doNotify notifiers
          end

    fun pReverseConnect thisNode toNode =
          let
            val revConn = pReverseConnections toNode
          in
            revConn := thisNode::(!revConn)
          end

    fun pReverseDisconnect thisNode toNode =
          let
            val revConn = pReverseConnections toNode
          in
            revConn := rev (foldl (pRemoveNode thisNode) nil (!revConn))
          end

    fun pDoConnect fc rc thisNode toNode =
          let
            val performNotify = fc thisNode toNode
            val _ = rc thisNode toNode
          in
            performNotify toNode
          end

    val connect = pDoConnect pConnect pReverseConnect
    val disconnect = pDoConnect pDisconnect pReverseDisconnect

    fun connected thisNode =
          let
            val (nodes, _) =
                   (pConnections thisNode)
          in
            map (fn (aNode, _) => aNode) (!nodes)
          end

    fun reverseConnected theNode =
          let
            val revConn = pReverseConnections theNode
          in
            !revConn
          end

    (* FIXME: Implement more efficiently with a dictionary *)
    fun pToType f thisNode theType =
          let
            val nodes = f thisNode
            (* FIXME: This would have been a constant time seach
                      with a dictionary implemented as a hashtable *)
            fun findIt (aNode, result) =
                  if typeName aNode = theType then
                    aNode::result
                  else result
          in
            foldl findIt nil nodes
          end

    val connectedToType =
          pToType connected

    val reverseConnectedFromType =
          pToType reverseConnected

    fun pSingleToType f msgConn thisNode theType =
          let
            val nodes = f thisNode theType
          in
            if length nodes = 1 then
              hd nodes
            (* FIXME: Make better error msg *)
            else (Err.error(msgConn ^ " from " ^
                  (typeName thisNode) ^ " to " ^
                  theType ^ " is not a single connection");
                  hd nodes) (* Notice an exception should be raised from Err.error *)
          end

    val singleConnectedToType =
          pSingleToType connectedToType "Connection"

    val singleReverseConnectedFromType =
          pSingleToType
            reverseConnectedFromType
            "Reverse connection"

    fun isConnected thisNode toNode =
          let
            val (nodes, _) =
                   pConnections thisNode
          
            fun contains nil = false
              | contains ((a, _)::rest) =
                  if a = toNode then true
                  else contains rest
          in
            contains (!nodes)
          end

    (* Notifications *)
    fun pAddNotify notify notifiers =
          let
            val ID = ref ()
            val _ = notifiers := (notify, ID)::(!notifiers)
          in
            ID
          end

    fun addConnectNotify thisNode notify =
          let
            val (_, notifiers) =
                   pConnections thisNode
          in
            pAddNotify notify notifiers
          end

    fun addDisconnectNotify thisNode removableNode notify =
          let
            val (nodes, _) =
                   pConnections thisNode
            val ID = ref ()
            fun addIt (aNode, notifiers) =
                  if aNode = removableNode then
                    (aNode, (notify, ID)::notifiers)
                  else
                    (aNode, notifiers)
            val _ = nodes := map addIt (!nodes)
          in
            ID
          end
  end

structure DGNodes = FDGNodes(ErrorStdOut)
