(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

functor FGenericLF() =
  struct
    fun drawBackground dp r c =
          let
            val _ = PIntDrawable.setFGColor dp c
            val _ = PIntDrawable.fillRect dp r
          in
            ()
          end

    fun lighten {r, g, b} =
          let
            fun l v =
                  if v + 0.3 > 1.0 then
                    1.0
                  else v + 0.3
          in
            {r = l r, g = l g, b = l b}
          end

    fun darken {r, g, b} =
          let
            fun d v =
                  if v - 0.3 < 0.0 then
                    0.0
                  else v - 0.3
          in
            {r = d r, g = d g, b = d b}
          end

    fun drawLitFrame dp {x, y, w, h} upperLeft lowerRight =
          let
            (* Lines are different to draw than rects... *)
            val w = w - 1
            val h = h - 1
            
            (* Upper left edges *)
            val _ = PIntDrawable.setFGColor dp upperLeft
            val _ = PIntDrawable.drawLine dp {x1 = x, y1 = y,
                                              x2 = x + w, y2 = y}
            val _ = PIntDrawable.drawLine dp {x1 = x, y1 = y,
                                              x2 = x, y2 = y + h}

            (* Lower right edges *)
            val _ = PIntDrawable.setFGColor dp lowerRight
            val _ = PIntDrawable.drawLine dp {x1 = x, y1 = y + h,
                                              x2 = x + w, y2 = y + h}
            val _ = PIntDrawable.drawLine dp {x1 = x + w, y1 = y,
                                              x2 = x + w, y2 = y + h}
          in
            ()
          end
  
    fun drawFrameOut dp r c =
          drawLitFrame dp r
            (lighten c)
            (darken c)

    fun drawFrameIn dp r c =
          drawLitFrame dp r
            (darken c)
            (lighten c)

    fun drawEditFrame f dp r c =
          (f dp r c;
           PIntDrawable.setFGColor dp {r = 0.0, g = 0.0, b = 0.0};
           PIntDrawable.drawRect dp (IRect2D.inset r 1))

    val drawEditFrameIn =
          drawEditFrame drawFrameIn

    val drawEditFrameOut =
          drawEditFrame drawFrameOut

    (* FIXME: XOR? Invert text? *)
    fun drawCursor dp r =
          drawBackground dp r
            {r = 1.0, g = 1.0, b = 1.0}

    fun clipLetters dp theText w =
          let
            val poss = PIntDrawable.charPositions dp theText
          
            fun clipIt (cPos::rest, index) =
                  if cPos > w then
                    Substring.string
                      (Substring.substring
                         (theText, 0, index))
                  else
                    clipIt (rest, index + 1)
              | clipIt (nil, _) = theText
          in
            clipIt (poss, 0)
          end

    fun centerX dp theText x w =
          let
            val textW = PIntDrawable.textWidth dp theText
          in
            x + (w div 2) - (textW div 2)
          end

    fun centerY dp theText y h =
          let
            val textH = PIntDrawable.textHeight dp theText
          in
            y + (h div 2) + (textH div 2)
          end

    fun drawCenteredText dp {x, y, w, h} theText =
          (PIntDrawable.setFGColor dp
             {r = 0.0, g = 0.0, b = 0.0};
           PIntDrawable.drawText dp
             {x = centerX dp theText x w, (* FIXME: Offset? *)
              y = centerY dp theText y h}
             theText)

    fun drawCenteredYText dp {x, y, w, h} theText =
          (PIntDrawable.setFGColor dp
             {r = 0.0, g = 0.0, b = 0.0};
           PIntDrawable.drawText dp
             {x = x, (* FIXME: Offset? *)
              y = centerY dp theText y h}
             (clipLetters dp theText w))
  end

structure GenericLF = FGenericLF()
