(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(*require "$.mlabstractuiwinnt.uiregister";*)
(*require "$.gtwinnt.dgshorthand";*)

functor FPortableListBox(Reg : UI_REGISTER
                         where type t = uiRegGUIClass.t) : UI_PLUGIN =
  struct
    val name = "ListBox"

    (* FIXME: We should access a global dictionary instead to
              get the user's look & feel... *)
    structure PortableScrollBar = FPortableScrollBar(DummyReg)
    
    fun construct () =
          let
            val textlines = "TextLines"
            val selectableItems = "SelectableItems"
            val doubleClick = "OnDoubleClick"
            val click = "OnClick"

            val (nameNode, _) = ParamName.create name

            (* FIXME: Notifications should really be done through the connections -
                      unless it is considered illegal for the user to connect
                      stuff to size, text, create, native etc. 
                      Maybe I should just add connection constraints?
                      Or notifications for stateinfo nodes? *)

            (* Add parameter nodes to the graph *)
            val textlinesProt =
                  DGPStrings.add nameNode textlines (nil, nameNode)
            val selectableItemsProt =
                  DGPItemSelections.add nameNode selectableItems nil
            val clickNProt =
                  DGPPointerButtonNotify.add nameNode click ()
            val doubleClickNProt =
                  DGPPointerButtonNotify.add nameNode doubleClick ()
            val hasFocusProt =
                  DGPBool.add nameNode "HasFocus" (false, nameNode)

            fun overrideTextLines invalidateF =
                  (* Notice: Actually we don't override -
                             we just add a notifier *)
                  PStrings.addChangedNotifier
                    textlinesProt (fn _ => invalidateF)

            fun overrideSelectableItems invalidateF =
                  let
                    val {count, selectAt, deselectAll, isSelectedAt} =
                           selectableItemsProt

                    fun doCount () =
                          PStrings.count textlinesProt

                    fun newSelectAt inh index =
                          if (PMethod.execute count) () <= index orelse
                             index < 0 then
                            Excn.eRaiseGeneric
                              ("RANGE",
                               "Index out of range " ^
                               "in PItemSelections.selectAt")
                          else if (!isSelectedAt) index then
                            ()
                          else
                            (inh index;
                             invalidateF ())
                  in
                    (PMethod.replace count doCount;
                     PMethod.override1Arg selectAt newSelectAt;
                     PMethod.wrap1Arg deselectAll invalidateF)
                  end

            fun genericConstruct {invalidateRect, getArea} =
                  let
                    val textTop = 4
                  
                    fun doRedraw dp rects =
                          let
                            val (r as {x, y, w, h}) =
                                   ParamRect.getRect nameNode
                            val textMargin = 4
                            val scrollBarWidth = 18
                            val height = PIntDrawable.textHeight dp "yX"
                            val linesCnt = PStrings.count textlinesProt

                            fun drawItem index yPos =
                                  let
                                    val r = {x = textMargin, y = yPos,
                                             w = w - textMargin - textMargin,
                                             h = height}

                                    val _ = if PItemSelections.isSelectedAt
                                                 selectableItemsProt
                                                 index then
                                              GenericLF.drawBackground dp r
                                                {r = 0.4, g = 0.4, b = 0.4}
                                            else ()
                                  in
                                    GenericLF.drawCenteredYText dp
                                      {x = textMargin, y = yPos,
                                       w = w - textMargin - textMargin,
                                       h = height}
                                      (PStrings.readAt textlinesProt index)
                                  end

                            fun drawItems index yPos =
                                  if index < linesCnt andalso
                                     yPos + height < h - textTop then
                                       (drawItem index yPos;
                                        drawItems (index + 1) (yPos + height))
                                  else ()

                            (* FIXME: Communicate with scrollbar control *)
                            (*fun calcRel space room occupied =
                                  if room > occupied then
                                    space
                                  else
                                    (space * room) div occupied

                            fun drawScrollBar r =
                                  let
                                    val {x, y, w, h} = IRect2D.inset r 2
                                    val relRect =
                                          {x = x, y = y, w = w,
                                           h = calcRel h (h div height) linesCnt}
                                  in
                                    (GenericLF.drawBackground dp r {r = 0.3, g = 0.4, b = 0.5};
                                     GenericLF.drawFrameIn dp r {r = 0.3, g = 0.4, b = 0.5};
                                     GenericLF.drawBackground dp relRect {r = 0.4, g = 0.4, b = 0.4};
                                     GenericLF.drawFrameOut dp relRect {r = 0.4, g = 0.4, b = 0.4})
                                  end *)

                            val _ = map (fn re =>
                                           GenericLF.drawBackground dp re
                                             {r = 0.5, g = 0.5, b = 0.5})
                                        rects
                          in
                            (GenericLF.drawEditFrameIn dp r {r = 0.3, g = 0.4, b = 0.5};
                             drawItems 0 textTop)
                          end

                    (* FIXME: Make this standard? *)
                    fun invalidate () =
                          let
                            val {w, h, x, y} = getArea ()
                          in
                            invalidateRect {x = 0, y = 0, w = w, h = h}
                          end

                    fun deselectAll () =
                          PItemSelections.deselectAll
                            selectableItemsProt

                    fun selectItemAt {x, y} =
                          let
                            val dp = ParamDrawableNode.getNativeProtocol nameNode
                            val height = PIntDrawable.textHeight dp "yX"
                            val index = if y >= textTop then
                                          (y - textTop) div height
                                        else ~1
                            val lines = PItemSelections.count
                                          selectableItemsProt
                          in
                            if index >= 0 andalso index < lines then
                              PItemSelections.selectAt
                                selectableItemsProt index
                            else ()
                          end

                    val selecting = ref false

                    fun doPointerDown {button, point} =
                          if button = 1 then
                            (deselectAll ();
                             selectItemAt point;
                             selecting := true)
                          else
                            ()

                    fun doPointerMove {point} =
                          if !selecting then
                            selectItemAt point
                          else ()

                    fun doPointerUp {button, point} =
                          if button = 1 then
                            selecting := false
                          else ()

                    val _ = overrideSelectableItems invalidate
                    val _ = overrideTextLines invalidate
                  in
                    {redraw = doRedraw,
                     pointerDown = doPointerDown,
                     pointerUp = doPointerUp,
                     pointerValue = (fn _ => ()),
                     pointerMove = doPointerMove,
                     keyPress = (fn _ => ())}
                  end
          
            val _ = GenericUIControl.constructInt
                      nameNode
                      {w = 150, h = 150}
                      genericConstruct

            (* Adding a scrollbar... *)
            open DGShorthand
            val scrollBar = PortableScrollBar.construct ()
            (* FIXME: Make an addChild method in LFNHierarchy... *)
            val children = nameNode +: "Children"
            val _ = children ~+ scrollBar

            fun onListboxSizeXChange sender value =
                  let
                    val scrollBarWidth =
                          PInt.read (DGPInt.p (scrollBar +: "SizeX"))
                  in
                    LFNGeometry.setPosX scrollBar
                      (value - scrollBarWidth)
                  end

            fun onListboxSizeYChange sender value =
                  LFNGeometry.setSizeY scrollBar value

            val sizeXNode = nameNode +: "SizeX"
            val sizeXProt = DGPInt.p sizeXNode
            val _ = PInt.addChangedNotifier sizeXProt
                      onListboxSizeXChange

            val sizeYNode = nameNode +: "SizeY"
            val sizeYProt = DGPInt.p sizeYNode
            val _ = PInt.addChangedNotifier sizeYProt
                      onListboxSizeYChange

            val _ = onListboxSizeXChange nameNode
                      (PInt.read (DGPInt.p (nameNode +: "SizeX")))

            fun onCreateControl sender value =
                  if value then
                    LFNMisc.createControl scrollBar
                  else ()
            val createNode = nameNode +: "Create"
            val createProt = DGPBool.p createNode
            val _ = PBool.addChangedNotifier createProt onCreateControl
          in
            nameNode
          end

    val _ = Reg.register name construct
  end
