(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

functor FPortableTextField(Reg : UI_REGISTER
                           where type t = uiRegGUIClass.t) : UI_PLUGIN =
  struct
    val name = "TextField"
    
    fun construct () =
          let
            val etext = "EditableText"

            val (nameNode, _) = ParamName.create name

            (* FIXME: Notifications should really be done through the connections -
                      unless it is considered illegal for the user to connect
                      stuff to size, text, create, native etc. 
                      Maybe I should just add connection constraints?
                      Or notifications for stateinfo nodes? *)

            
            (* Add parameter nodes to the graph *)
            val etextProt =
                  DGPString.add nameNode etext ("", nameNode)
            val hasFocusProt =
                  DGPBool.add nameNode "HasFocus" (false, nameNode)

            fun overrideHasFocus invalidateF cursorPos =
                  let
                    fun newWriteHasFocus inherited value =
                          let
                            (* Set focusControlNode...
                               Watch out for infinite mutual recursion! *)
                            fun doWrite value =
                                  let
                                    val topNode =
                                          LFNHierarchy.firstParentOfType
                                            nameNode "Window"
                                  in
                                    LFNMisc.setFocusControlNode
                                      topNode value
                                  end
                          in
                            if PBool.read hasFocusProt = value then
                              ()
                            else
                              let
                                val maxCursor = String.size (PString.read etextProt)
                              in
                                (inherited value;(* First store new value *)
                                 case value of
                                   true =>
                                     (if maxCursor < (!cursorPos) then
                                        cursorPos := maxCursor
                                      else ();
                                      doWrite (SOME(nameNode)))
                                 | false =>
                                     doWrite NONE;
                                 invalidateF ())
                              end
                          end
                    
                    val {write = writeHasFocus, ...} = hasFocusProt
                    (* FIXME: ? To make sure that the owner
                              window is told if we get focus... *)
                    val _ = PBool.write hasFocusProt false
                  in
                    PMethod.override1Arg
                       writeHasFocus
                       newWriteHasFocus
                  end

            fun overrideEText invalidateF cursorPos =
                  let
                    fun newWriteEText inh value =
                          if value = PString.read etextProt then
                            inh value
                          else
                            (* FIXME: Make this better/more efficient? *)
                            (if String.size value < (!cursorPos) then
                               PBool.write hasFocusProt false
                                 (* to avoid cursor out of range *)
                             else ();
                             inh value;
                             invalidateF ())

                    val {write = writeEText, ...} = etextProt
                  in
                    PMethod.override1Arg writeEText newWriteEText
                  end

            fun genericConstruct {invalidateRect, getArea} =
                  let
                    val cursorPos = ref 0
                    val cursorStart = 4

                    fun calcCursorX dp cp eText =
                          let
                            val s = Substring.substring
                                      (eText, 0, cp)
                          in
                            PIntDrawable.textWidth dp
                              (Substring.string s)
                          end
                  
                    fun doRedraw dp rects =
                          let
                            val r = ParamRect.getRect nameNode
                            val (r4 as {x, y, w, h}) = IRect2D.inset r 4
                            val cp = !cursorPos (* FIXME: It's not even thread safe when reading the cursorPos before eText (example: erase char... *)
                            val eText = PString.read etextProt

                            val _ = map (fn re =>
                                           GenericLF.drawBackground dp re
                                             {r = 0.5, g = 0.5, b = 0.5})
                                        rects
                            val _ = GenericLF.drawEditFrameIn dp r {r = 0.3, g = 0.4, b = 0.5}
                            val _ = GenericLF.drawCenteredYText dp r4 eText
                          in
                            (* FIXME: Invert text? XOR? *)
                            if PBool.read hasFocusProt then
                              GenericLF.drawCursor dp
                                {x = (calcCursorX dp cp eText) + x,
                                 y = y, w = 2, h = h}
                            else
                              ()
                          end

                    (* FIXME: Make this standard? *)
                    fun invalidate () =
                          let
                            val {w, h, x, y} = getArea ()
                          in
                            invalidateRect {x = 0, y = 0, w = w, h = h}
                          end

                    fun placeCursor {x, y} =
                          let
                            val ss = Substring.all (PString.read etextProt)

                            val placed = ref false
                            val px = x - cursorStart
                            val dp = ParamDrawableNode.getNativeProtocol nameNode

                            fun checkPlace () =
                                  if !placed then ()
                                  else
                                    cursorPos := Substring.size ss

                            fun placeClosest d1 d2 index =
                                  (if d1 < (~d2) then
                                     cursorPos := index
                                   else
                                     cursorPos := index + 1;
                                   placed := true)

                            fun foldPlace (c, (s, d, i)) =
                                  if !placed then (s, d, i)
                                  else
                                    let
                                      val newS = s ^ (Char.toString c)
                                      val newDist =
                                            px - (PIntDrawable.textWidth
                                                    dp newS)
                                    in
                                      if newDist < 0 then
                                        (placeClosest d newDist i;
                                         (s, d, i))
                                      else
                                        (newS, newDist, i + 1)
                                    end
                          in
                            if px <= 0 then
                              cursorPos := 0
                            else
                              (Substring.foldl foldPlace ("", px, 0) ss;
                               checkPlace ())
                          end

                    fun pointerDown {button, point} =
                          (PBool.write hasFocusProt true;
                           placeCursor point;
                           invalidate ())

                    (* f : substring -> string *)
                    fun updateEText f afterF =
                          (PString.write etextProt
                              (f (Substring.all (PString.read etextProt)));
                           afterF ())

                    fun moveCursor towardsRight =
                          cursorPos := (!cursorPos) + towardsRight

                    fun boundedMoveCursor towardsRight =
                          let
                            val newPos = (!cursorPos) + towardsRight
                            val sz = String.size (PString.read etextProt)
                          in
                            if newPos < 0 then
                              cursorPos := 0
                            else if newPos > sz then
                              cursorPos := sz
                            else
                              cursorPos := newPos
                          end

                    (* FIXME: Portable??? *)
                    fun isErase c =
                          (c = #"\^H")
                    fun isKill c =
                          (c = #"\^X")

                    fun insertEText inString oldSS =
                          let
                            val (left, right) =
                                   Substring.splitAt
                                     (oldSS, !cursorPos)
                            val inSS = Substring.all inString

                            (* It's not safe to set cursor before updating eText *)
                            (*val _ = moveCursor 1*)
                          in
                            Substring.concat
                              [left, inSS, right]
                          end

                    fun eraseETextChar oldSS =
                          let
                            val (left, right) =
                                   Substring.splitAt
                                     (oldSS, !cursorPos)
                            val newLeft =
                                  Substring.slice
                                    (left, 0, SOME((Substring.size left) - 1))
                            val _ = moveCursor (~1)
                          in
                            Substring.concat
                              [newLeft, right]
                          end

                    fun handleChar c =
                          if Char.isPrint c then
                            updateEText
                              (insertEText (Char.toString c))
                              (fn () => moveCursor 1)
                          else if isErase c andalso
                                  (!cursorPos) > 0 then
                            updateEText eraseETextChar (fn () => ())
                          else ()

                    (* FIXME: Portable??? *)
                    fun handleCode 147 =
                          boundedMoveCursor 1 (* Arrow right *)
                      | handleCode 148 =
                          boundedMoveCursor (~1) (* Arrow left *)
                      | handleCode _ = ()

                    fun keyPress {symbol} = (* FIXME: symbol *)
                          (case symbol of
                             GenUI.KeyString(sym) =>
                               (Substring.app
                                  handleChar
                                  (Substring.all sym))
                           | GenUI.KeyCode(code) =>
                               (handleCode code);
                           invalidate ()) (* FIXME: Only invalidate if necessary... *)

                    val _ = overrideHasFocus invalidate cursorPos
                    val _ = overrideEText invalidate cursorPos
                  in
                    {redraw = doRedraw,
                     pointerDown = pointerDown,
                     pointerUp = (fn _ => ()),
                     pointerValue = (fn _ => ()),
                     pointerMove = (fn _ => ()),
                     keyPress = keyPress}
                  end
                  
          
            val _ = GenericUIControl.constructInt
                      nameNode
                      {w = 100, h = 18}
                      genericConstruct
          in
            nameNode
          end

    val _ = Reg.register name construct
  end
