(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(*structure PEXeneGeneric = PFEXeneNative(type nativeType = (EXeneWin.window * Interact.in_env)
                                        val nativeName = "EXene generic")
structure DGPEXeneGeneric = FDGProtocols(PEXeneGeneric)

functor FEXeneGenericUtil() =
  struct
    fun create native =
          let
            val protocol = PEXeneGeneric.P.create
                             {readNative = (fn () => native)}
          in
            DGNodes.create "EXene generic" protocol
          end

    fun getNativeProtocol graphNode =
          let
            val nativeNode = DGNodes.singleConnectedToType
                               graphNode "EXene generic"
          in
            DGPEXeneGeneric.p nativeNode
          end

    fun getNativeWindow graphNode =
          PEXeneGeneric.readNative (getNativeProtocol graphNode)
  end

structure EXeneGenericUtil = FEXeneGenericUtil() *)

functor FEXeneGeneric() : GENERIC_UI_CONTROL =
  struct
    structure DR = Drawing
    structure EW = EXeneWin
    structure EXB = EXeneBase
    structure Geom = Geometry
    structure Inte = Interact

    val name = "Generic"

    fun constructInt nameNode {w, h} basicF =
          let
            (* FIXME: Notifications should really be done through the connections -
                      unless it is considered illegal for the user to connect
                      stuff to size, text, create, native etc. 
                      Maybe I should just add connection constraints?
                      Or notifications for stateinfo nodes? *)

            (* Add position and size parameter nodes to the graph *)
            val {protPosX, protPosY, protSizeX, protSizeY,
                 getArea} =
                   ParamRect.addRect nameNode
                     {x = 0, y = 0, w = w, h = h}
            val visibilityProt = DGPBool.add nameNode "Visible" (true, nameNode)
            val pointerDownNProt =
                  DGPPointerButtonNotify.add
                    nameNode "OnPointerDown" ()

            (* Add eXene native information to the graph *)
            (* FIXME: Move the registerChild, invalidateRect and getOffsetTotal
                      to the pass-along parameters in the generic UI-control *)
            fun createNativeInfo window inEnv
                                 registerChild invalidateRect
                                 getOffsetTotal graphNode =
                  let
                    val nativeNode = EXeneWindowUtil.create
                                       (window, inEnv, registerChild,
                                        invalidateRect, getOffsetTotal)
                    val _ = DGNodes.connect graphNode nativeNode
                  in
                    DGPEXeneWindow.p nativeNode
                  end

            fun overrideGeom invalidateRect =
                  let
                    (* Notice that this invalidate rect is without offset...
                       maybe this code belongs in the control's parent? *)
                    fun min (a, b) = if a < b then a else b
                    fun max (a, b) = if a > b then a else b
                    fun diff (a, b) = if a < b then b - a else a - b

                    (* FIXME: Some higher order magic and/or a few extra
                              functions in IRect2D would be good here... *)
                    fun newWritePosX inh value =
                          let
                            val oldValue = PInt.read protPosX
                          in
                            if value = oldValue then
                              inh value
                            else
                              (inh value;
                               invalidateRect {x = min (value, oldValue),
                                               y = PInt.read protPosY,
                                               w = PInt.read protSizeX + diff (value, oldValue),
                                               h = PInt.read protSizeY})
                          end

                    fun newWritePosY inh value =
                          let
                            val oldValue = PInt.read protPosY
                          in
                            if value = oldValue then
                              inh value
                            else
                              (inh value;
                               (* FIXME: Invalidate 2 rects instead? *)
                               invalidateRect {y = min (value, oldValue),
                                               x = PInt.read protPosX,
                                               h = PInt.read protSizeY + diff (value, oldValue),
                                               w = PInt.read protSizeX})
                          end

                    fun newWriteSizeX inh value =
                          let
                            val {x, y, w, h} = getArea ()
                          in
                            if value = w then
                              inh value
                            else
                              (inh value;
                                (* FIXME: Invalidate 2 rects instead? *)
                               invalidateRect {x = x, y = y,
                                               w = max (value, w), h = h})
                          end

                    fun newWriteSizeY inh value =
                          let
                            val {x, y, w, h} = getArea ()
                          in
                            if value = h then
                              inh value
                            else
                              (inh value;
                               invalidateRect {x = x, y = y,
                                               w = w, h = max (value, h)})
                          end

                    val {write = writePosX, ...} = protPosX
                    val {write = writePosY, ...} = protPosY
                    val {write = writeSizeX, ...} = protSizeX
                    val {write = writeSizeY, ...} = protSizeY
                  in
                    (PMethod.override1Arg writePosX newWritePosX;
                     PMethod.override1Arg writePosY newWritePosY;
                     PMethod.override1Arg writeSizeX newWriteSizeX;
                     PMethod.override1Arg writeSizeY newWriteSizeY)
                  end

            fun nativeCreate () =
                    let
                      val {x, y, w, h} = getArea ()

                      val parentNode = LFNHierarchy.parentNode nameNode
                      (* FIXME: RegisterChildAtParent, invalidateRect and
                                getOffsetParent should be moved to the
                                pass-on parameters...
                                So parentNode above should really be
                                firstParentOfType... *)
                      val (window, inEnv, registerChildAtParent,
                           invalidateRect, getOffsetParent) =
                             EXeneWindowUtil.getNativeWindow parentNode
                      val screen = EW.screenOfWin window
                      val display = EW.displayOfWin window

                      val {children, registerChild} =
                             HierarchyCore.prepareForChildren ()

                      (* No offset on the top-level window... *)
                      fun getOffsetTotal () =
                            let
                              val {x = offPX, y = offPY} = getOffsetParent ()
                              val {x = offMyX, y = offMyY, ...} = getArea ()
                            in
                              {x = offMyX + offPX, y = offMyY + offPY}
                            end

                      (* Calculate offset for drawable... *)
                      (* FIXME: Also in EXeneWindow - make more general *)
                      fun getDrawableArea () =
                            let
                              val {w, h, ...} = getArea ()
                              val {x = offX, y = offY} = getOffsetTotal ()
                            in
                              {x = offX, y = offY, w = w, h = h}
                            end

                      val dp = EXeneDrawableNode.add nameNode
                                 (DR.drawableOfWin window)
                                 display screen getDrawableArea

                      fun doInvalidateRect {x, y, w, h} =
                            let
                              (* FIXME: Now we don't need to accumulate
                                 invalidateRect calls do we?
                                 Make this and getOffsetTotal more general *)
                              val {x = offsetX, y = offsetY, ...} = getArea ()
                            in
                              invalidateRect
                                {x = x + offsetX, y = y + offsetY,
                                 w = w, h = h}
                            end
                              

                      val {redraw, pointerDown, pointerUp,
                           pointerValue, pointerMove, keyPress} =
                             basicF {invalidateRect = doInvalidateRect,
                                     getArea = getArea}

                      fun myPointerDown param =
                            PPointerButtonNotify.doNotify
                              pointerDownNProt nameNode param

                      (* All widgets can assume a blank background *)
                      fun drawBackground rects =
                            map (fn r =>
                                   GenericLF.drawBackground dp r
                                     {r = 0.3, g = 0.4, b = 0.5})
                                rects

                      fun redrawControl rects =
                            (drawBackground rects;
                             redraw dp rects) 

                      fun doRedraw rects =
                            HierarchyCore.handleDraw
                              redrawControl
                              (!children) rects

                      fun doPointerDown param =
                            (pointerDown param;
                             myPointerDown param)

                      val _ = registerChildAtParent
                                 {node = nameNode,
                                  getArea = getArea,
                                  redraw = doRedraw,
                                  pointerDown = doPointerDown,
                                  pointerUp = pointerUp,
                                  pointerValue = pointerValue,
                                  pointerMove = pointerMove,
                                  keyPress = keyPress}

                      fun textChanged value =
                            ()(* FIXME: uiWinUtil.setText win value *)

                      (* Override writeText... *)
                      (*val _ = writeText := textChanged*)

                      (* FIXME: Event handler on close *)
                      (*val _ = w.addMessageHandler
                                  (nativeHandle, w.WM_CLOSE,
                                   fn _ => (w.destroyWindow nativeHandle;
                                            raise Excn.eRaiseGeneric
                                              ("AUI_QUIT", "Application quitted!");
                                            SOME w.nullWord))*)
                    
                    val _ = createNativeInfo
                              window inEnv
                              registerChild
                              doInvalidateRect
                              getOffsetTotal
                              nameNode

                    val _ = doInvalidateRect
                              {x = 0, y = 0, w = w, h = h}

                    (* Notice that this invalidate rect is without offset...
                       maybe this code belongs in the control's parent? *)
                    val _ = overrideGeom invalidateRect
                  in
                    ()
                  end

            val _ = ParamCreate.setup nameNode nativeCreate (fn _ => ())
          in
            ()
          end
  end
