(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure PEXeneWindow = PFEXeneNative(type nativeType = (EXeneWin.window * Interact.in_env *
                                                            (GenUI.bInternal -> unit) *
                                                            (IRect2D.t -> unit) *
                                                            (unit -> IPoint2D.t))
                                         val nativeName = "EXene window")
structure DGPEXeneWindow = FDGProtocols(PEXeneWindow)

functor FEXeneWindowUtil() =
  struct
    fun create native =
          let
            val protocol = PEXeneWindow.P.create
                             {readNative = (fn () => native)}
          in
            DGNodes.create "EXene window" protocol
          end

    fun getNativeProtocol graphNode =
          let
            val nativeNode = DGNodes.singleConnectedToType
                               graphNode "EXene window"
          in
            DGPEXeneWindow.p nativeNode
          end

    fun getNativeWindow graphNode =
          PEXeneWindow.readNative (getNativeProtocol graphNode)
  end

structure EXeneWindowUtil = FEXeneWindowUtil()

functor FEXeneWindow(Reg : UI_REGISTER
                     where type t = uiRegGUIClass.t) : UI_PLUGIN =
  struct
    structure EW = EXeneWin
    structure EXB = EXeneBase
    structure Geom = Geometry
    structure Inte = Interact
    structure DR = Drawing
    val name = "Window"

    fun construct () =
          let
            val text = "Text"

            val (nameNode, _) = ParamName.create name

            (* FIXME: Notifications should really be done through the connections -
                      unless it is considered illegal for the user to connect
                      stuff to size, text, create, native etc. 
                      Maybe I should just add connection constraints?
                      Or notifications for stateinfo nodes? *)

            (* Add position and size parameter nodes to the graph *)
            val {protPosX, protPosY, protSizeX, protSizeY,
                 getArea = getWindowArea} =
                   ParamRect.addRect nameNode
                     {x = 0, y = 0, w = 400, h = 400}
            val visibilityProt = DGPBool.add nameNode "Visible" (true, nameNode)
            val textProt =
                  DGPString.add nameNode "Text" ("", nameNode)
            val focusControlProt =
                  DGPDGNode.add nameNode "FocusControl" (NONE, nameNode)
            val pointerDownNProt =
                  DGPPointerButtonNotify.add
                    nameNode "OnPointerDown" ()

            (* Add eXene native information to the graph *)
            (* FIXME: Move the registerChild, invalidateRect and getOffsetTotal
                      to the pass-along parameters in the generic UI-control *)
            fun createNativeInfo window inEnv
                                 registerChild invalidateRect
                                 getOffsetTotal graphNode =
                  let
                    val nativeNode = EXeneWindowUtil.create
                                       (window, inEnv, registerChild,
                                        invalidateRect, getOffsetTotal)
                    val _ = DGNodes.connect graphNode nativeNode
                  in
                    DGPEXeneWindow.p nativeNode
                  end

            fun nativeCreate () =
                    let
                      val {x, y, w, h} = getWindowArea ()

                      val masterNode = LFNHierarchy.parentNode nameNode
                      val screen = EXeneDisplay.getNativeScreen masterNode

                      val winSize = Geom.SIZE{wid=w, ht=h}
                      val minSize = Geom.SIZE{wid=0, ht=0} (* FIXME: ... *)
                      val winPos = Geom.PT{x = x, y = y}
	              val (win, inEnv) =
                            EW.createSimpleTopWin screen
                              {geom = Geom.WGEOM{pos=winPos, sz=winSize, border=1},
		               border = EXeneColor.toEXene screen {r = 0.0, g = 0.0, b = 0.0},
		               backgrnd = EXeneColor.toEXene screen {r = 0.3, g = 0.4, b = 0.5}}
                      val display = EW.displayOfWin win

                      fun changeFocus value NONE = ()
                        | changeFocus value (SOME(aControl)) =
                            let
                              val valueNodes = DGNodes.connected aControl

                              fun changeIt valueNode =
                                    if DGNodes.typeName valueNode = "HasFocus" then
                                      PBool.write (DGPBool.p valueNode) value
                                    else ()
                            in
                              List.app changeIt valueNodes
                            end

                      fun newWriteFocusControl inh NONE =
                            (changeFocus false (PDGNode.read focusControlProt);
                             inh NONE)
                        | newWriteFocusControl inh (SOME(newControl)) =
                            (changeFocus false (PDGNode.read focusControlProt);
                             inh (SOME(newControl));
                             changeFocus true (SOME(newControl)))

                      val {write = writeFocusControl, ...} =
                             focusControlProt

                      val _ = PMethod.override1Arg
                                writeFocusControl
                                newWriteFocusControl
                      
                      fun textChanged value =
                            EXeneWin.setWMProperties win
                                {argv = SMLofNJ.getArgs(),
                                 win_name = SOME value,
                                 icon_name = SOME value,
                                 size_hints =
                                   [ICCC.HINT_PPosition,
                                    ICCC.HINT_PSize,
                                    ICCC.HINT_PMinSize minSize],
                                 wm_hints = [],
                                 class_hints = SOME{res_name= name, (* name is "Window" here *)
                                                    res_class= name}
                                }

                      val _ = textChanged (PString.read textProt)

                      fun fromEXene (Geom.RECT{x, y, wid, ht}) =
                            {x = x, y = y, w = wid, h = ht}

                      (* NOTICE: Not used... *)
                      fun toEXene {x, y, w, h} =
                            Geom.RECT{x = x, y = y, wid = w, ht = h}

                      val {children, registerChild} =
                             HierarchyCore.prepareForChildren ()

                      (* No offset on the top-level window... *)
                      fun getOffsetTotal () = {x = 0, y = 0}

                      (* Calculate offset for drawable... *)
                      (* FIXME: Also in EXeneGeneric - make it more general *)
                      fun getDrawableArea () =
                            let
                              val {x, y, w, h} = getWindowArea ()
                              val {x = offX, y = offY} = getOffsetTotal ()
                            in
                              {x = offX, y = offY, w = w, h = h}
                            end

                      val dp = EXeneDrawableNode.add nameNode
                                 (DR.drawableOfWin win)
                                 display screen getDrawableArea

                      (* All widgets can assume a blank background *)
                      fun myRedraw rects =
                            map (fn r =>
                                   GenericLF.drawBackground dp r
                                     {r = 0.3, g = 0.4, b = 0.5})
                                rects

                      val {write = writePosX, ...} = protPosX
                      val {write = writePosY, ...} = protPosY
                      val {write = writeSizeX, ...} = protSizeX
                      val {write = writeSizeY, ...} = protSizeY
                      val oldWritePosX = !writePosX
                      val oldWritePosY = !writePosY
                      val oldWriteSizeX = !writeSizeX
                      val oldWriteSizeY = !writeSizeY

                      (* FIXME: Make an area node which implements
                                the PosX, PosY, SizeX and SizeY nodes...?
                                But then there's the override issue... *)
                      fun newWritePosX inherited value =
                            (EW.moveWin win
                               (Geom.PT{x = value,
                                        y = PInt.read protPosY});
                             inherited value)

                      fun newWritePosY inherited value =
                            (EW.moveWin win
                               (Geom.PT{x = PInt.read protPosX,
                                        y = value});
                             inherited value)

                      fun newWriteSizeX inherited value =
                            (EW.resizeWin win
                               (Geom.SIZE{wid = value,
                                          ht = PInt.read protSizeY});
                             inherited value)

                      fun newWriteSizeY inherited value =
                            (EW.resizeWin win
                               (Geom.SIZE{wid = PInt.read protSizeX,
                                          ht = value});
                             inherited value)

                      val _ = PMethod.override1Arg
                                writePosX
                                newWritePosX
                      
                      val _ = PMethod.override1Arg
                                writePosY
                                newWritePosY
                      
                      val _ = PMethod.override1Arg
                                writeSizeX
                                newWriteSizeX
                      
                      val _ = PMethod.override1Arg
                                writeSizeY
                                newWriteSizeY

                      fun handleCmd (Inte.CI_Redraw(_)) =
                            Excn.eRaiseBug "Redraw should not be sent to command thread!"
                        | handleCmd (Inte.CI_Resize(rect)) =
                            let
                              val {x, y, w, h} = fromEXene rect
                            in
                              (oldWritePosX x; oldWritePosY y;
                               oldWriteSizeX w; oldWriteSizeY h)
                            end
                        | handleCmd (Inte.CI_ChildBirth(win)) =
                            print "Window child birth "
                        | handleCmd (Inte.CI_ChildDeath(win)) =
                            print "Window child death "
                        | handleCmd (Inte.CI_OwnDeath) =
                            print "Window own death "

                      fun myPointerDown param =
                            PPointerButtonNotify.doNotify
                              pointerDownNProt nameNode param

                      fun myPointerUp _ = ()

                      fun myPointerValue _ = ()

                      fun myPointerMove _ = ()

                      datatype inputMsg =
                        InKey of Inte.kbd_msg
                      | InMouse of Inte.mouse_msg

                      fun getGetArea
                            ({getArea, ...} : GenUI.bInternal) =
                              getArea

                      fun getNode
                            ({node, ...} : GenUI.bInternal) =
                              node

                      fun getPointerDown
                            ({pointerDown, ...} : GenUI.bInternal) =
                              pointerDown

                      fun getPointerUp
                            ({pointerUp, ...} : GenUI.bInternal) =
                              pointerUp

                      fun getPointerMove
                            ({pointerMove, ...} : GenUI.bInternal) =
                              pointerMove

                      fun getPointerValue
                            ({pointerValue, ...} : GenUI.bInternal) =
                              pointerValue

                      fun checkHit (thePoint as {x, y}) theChild hitF =
                            let
                              val (area as {x = offsetX,
                                            y = offsetY, ...}) =
                                     getGetArea theChild ()
                              
                              val newPoint = {x = x - offsetX,
                                              y = y - offsetY}
                            in
                              if IRect2D.containsPoint area thePoint then
                                (hitF newPoint; true)
                              else
                                false
                            end

                      fun checkDispatch checkDispatchChild dispatchOtherwise =
                            let
                              fun checkIt nil =
                                    dispatchOtherwise ()
                                | checkIt (aChild::rest) =
                                    if checkDispatchChild aChild then
                                      ()
                                    else
                                      checkIt rest
                            in
                              checkIt (!children)
                            end

                      fun doDispatch paramToPoint replaceParamPoint
                                     getF otherwise param =
                            let
                              fun checkChild theChild = 
                                    checkHit
                                       (paramToPoint param)
                                       theChild
                                       (fn newPoint =>
                                          (getF theChild)
                                          (replaceParamPoint param newPoint))
                            in
                              checkDispatch
                                checkChild
                                (fn () => otherwise param)
                            end

                      fun generalPointerValue otherwise param =
                            doDispatch
                              (fn {button, point, value} => point)
                              (fn {button, point, value} =>
                                 fn newPoint =>
                                   {button = button,
                                    point = newPoint,
                                    value = value})
                              getPointerValue 
                              otherwise
                              param

                      fun generalPointerUpDown getF otherwise param =
                            doDispatch
                              (fn {button, point} => point)
                              (fn {button, point} =>
                                 fn newPoint =>
                                   {button = button,
                                    point = newPoint})
                              getF 
                              otherwise
                              param

                      val generalPointerUp =
                            generalPointerUpDown getPointerUp
                      val generalPointerDown =
                            generalPointerUpDown getPointerDown

                      fun generalPointerMove otherwise param =
                            doDispatch
                              (fn {point} => point)
                              (fn {point} =>
                                 fn newPoint =>
                                   {point = newPoint})
                              getPointerMove 
                              otherwise
                              param

                      val doPointerValue =
                            generalPointerValue
                              myPointerValue

                      val doPointerDown =
                            generalPointerDown
                              myPointerDown

                      val doPointerUp =
                            generalPointerUp
                              myPointerUp

                      val doPointerMove =
                            generalPointerMove
                              myPointerMove
                      
                      fun handleMouseUpDown doPointerF value
                              {but, pt, scr_pt(* FIXME: , state*), time} =
                            let
                              val Inte.MButton(button) = but
                              val Geom.PT{x, y} = pt
                              val thePoint = {x = x, y = y}
                              (* FIXME: Add time... *)

                              val _ = doPointerValue
                                        {button = button,
                                         value = value,
                                         point = thePoint}
                              val _ = doPointerF
                                        {button = button,
                                         point = thePoint}
                            in () end

                      val handleMouseDown =
                            handleMouseUpDown doPointerDown 1.0
                      val handleMouseUp =
                            handleMouseUpDown doPointerUp 0.0

                      fun handleMouseMotion
                              {pt, scr_pt(* FIXME: , state*), time} =
                            let
                              val Geom.PT{x, y} = pt
                              val thePoint = {x = x, y = y}
                              (* FIXME: Add time... *)
                            in
                              doPointerMove
                                {point = thePoint}
                            end

                      fun handleKeyPress {symbol} =
                            let
                              (* FIXME: Should I pass the event
                                 all the way down the child chain
                                 rather than sending directly
                                 to the inner focused control? *)
                              fun doPress control ({node, keyPress, ...} : GenUI.bInternal) =
                                    if node = control then
                                      keyPress {symbol = symbol}
                                    else ()
                            in
                              case PDGNode.read focusControlProt of
                                      NONE => ()
                                    | SOME(control) =>
                                        List.app (doPress control) (!children)
                            end

                      fun defLookup keysym modkeyState =
                            Inte.lookupString
                              Inte.defaultTranslation
                                (keysym, modkeyState)

                      fun mylookup keysym modkeyState =
                            case keysym of
                               Inte.KEYSYM(sym) =>
                                 if sym = 65363 then (* Arrow right *)
                                   GenUI.KeyCode 147 (* FIXME: ... *)
                                 else if sym = 65361 then (* Arrow left *)
                                   GenUI.KeyCode 148 (* FIXME: ... *)
                                 else
                                   GenUI.KeyString(
                                     defLookup keysym modkeyState)
                             | _ =>
                                   GenUI.KeyString(
                                     defLookup keysym modkeyState)

                      fun handleInput (InKey(Inte.KEY_Press(keysym, modkeyState))) =
                            ((handleKeyPress
                                 {symbol = mylookup keysym modkeyState})
                             handle KeysymNotFound => ())
                                
                        | handleInput (InKey(_)) = ()
                        | handleInput (InMouse(Inte.MOUSE_Down
                                         {but, pt, scr_pt, state, time})) =
                            handleMouseDown
                              {but = but, pt = pt, scr_pt = scr_pt, time = time}
                        | handleInput (InMouse(Inte.MOUSE_FirstDown
                                         {but, pt, scr_pt, time})) =
                            handleMouseDown
                              {but = but, pt = pt, scr_pt = scr_pt, time = time}
                        | handleInput (InMouse(Inte.MOUSE_Up
                                         {but, pt, scr_pt, state, time})) =
                            handleMouseUp
                              {but = but, pt = pt, scr_pt = scr_pt, time = time}
                        | handleInput (InMouse(Inte.MOUSE_LastUp
                                         {but, pt, scr_pt, time})) =
                            handleMouseUp
                              {but = but, pt = pt, scr_pt = scr_pt, time = time}
                        | handleInput (InMouse(Inte.MOUSE_Motion
                                         {pt, scr_pt, time})) =
                            handleMouseMotion
                              {pt = pt, scr_pt = scr_pt, time = time}
                        | handleInput (InMouse(_)) = ()
                            

                      (* Create asynchronous communication channels *)
                      val inputMb = Mailbox.mailbox()
                      val drawMb = Mailbox.mailbox ()
                      val cmdMb = Mailbox.mailbox()

                      (* Setup dispatch thread to avoid locking up
                         input thread when handling the drawing etc. *)
                      val Inte.InEnv{k, m, ci, ...} = inEnv

                      fun invalidateRect rect =
                            Mailbox.send(drawMb, [rect])

                      (* Dispatchers *)
                      fun dispCmd msg =
                            let
                              val msgBody = Inte.msgBodyOf msg
                            in
                              case msgBody of
                                (Inte.CI_Redraw(rects)) =>
                                  Mailbox.send(drawMb, map fromEXene rects)
                              | _ =>
                                  Mailbox.send(cmdMb, msgBody)
                            end

                      fun dispMouse msg =
                            Mailbox.send(inputMb, InMouse(Inte.msgBodyOf msg))
                      
                      fun dispKey msg =
                            Mailbox.send(inputMb, InKey(Inte.msgBodyOf msg))

                      val cmdEvt = CML.wrap (ci, dispCmd)
                      val keyEvt = CML.wrap (k, dispKey)
                      val mouseEvt = CML.wrap (m, dispMouse)

                      fun dispLoop () =
                            (CML.select [cmdEvt, keyEvt, mouseEvt];
                             dispLoop ())
                            
                      val _ = CML.spawn dispLoop

                      (* Setup drawing thread - merge all redraw rects *)
                      fun checkDrawMore rects =
                            case Mailbox.recvPoll drawMb of
                              NONE =>
                                (HierarchyCore.handleDraw
                                   myRedraw (!children) rects;
                                 drawLoop ())
                            | SOME(moreRects) =>
                                checkDrawMore (moreRects@rects)
                      and drawLoop () =
                            checkDrawMore (Mailbox.recv drawMb)
                            
                      val _ = CML.spawn drawLoop

                      (* Setup command thread *)
                      fun cmdLoop () =
                            (handleCmd(Mailbox.recv cmdMb);
                             cmdLoop ())
                            
                      val _ = CML.spawn cmdLoop

                      (* Setup input thread *)
                      fun inputLoop () =
                            (handleInput(Mailbox.recv inputMb);
                             inputLoop ())
                      
                      val _ = CML.spawn inputLoop

                      (* Override writeText... *)
                      val {write = writeText, ...} = textProt

                      (* FIXME: Override read to read from eXene window rather
                                than just reading from own stored data? *)
                      val _ = PMethod.wrap1Arg
                                writeText
                                (fn () => textChanged
                                   (PString.read textProt))

                      (* FIXME: Is this necessary here? *)
	              val _ = EXeneWin.mapWin win

                      (* FIXME: Event handler on close *)
                      (*val _ = w.addMessageHandler
                                  (nativeHandle, w.WM_CLOSE,
                                   fn _ => (w.destroyWindow nativeHandle;
                                            raise Excn.eRaiseGeneric
                                              ("AUI_QUIT", "Application quitted!");
                                            SOME w.nullWord))*)

                    val _ = createNativeInfo
                              win inEnv
                              registerChild
                              invalidateRect
                              getOffsetTotal
                              nameNode

                    val _ = invalidateRect
                              {x = 0, y = 0, w = w, h = h}
                  in
                    ()
                  end

            val _ = ParamCreate.setup nameNode nativeCreate (fn _ => ())
          in
            nameNode
          end

    val _ = Reg.register name construct
  end
