(*
    This file is part of ML::AbstractUI - an abstract user interface
    API with a crossplatform implementation in Standard ML.

    Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(*require "typeneutral";*)
(*require "$.commonfunctionswinnt.util";*)
(*require "$.errorhandlingwinnt.exceptions";*)

(* This sourcecode file implements a general typeneutral class system. *)
(* It is used for implementing typeneutral protocols in ML::AbstractUI *)
(* Notice: This file is small but the code is complex! *)

(* CLASSES_IMPL and ClassesImpl has been defined just to hide the
   input and output types so that the user cannot execute the
   internal functions for handling typeneutralness except by
   using tryAssignObject *)
(* This is used only internally *)
signature CLASSES_IMPL =
  sig
    type object

    val createObject : TypeNeutral.setF -> string ref -> object
    
    val tryAssignObject : object -> TypeNeutral.getF -> bool

    val className : object -> string
  end

(* This is used only internally *)
structure ClassesImpl :> CLASSES_IMPL =
  struct
    type object = TypeNeutral.setF * string ref

    fun createObject setFnc classNameRef =
          (setFnc, classNameRef)

    fun tryAssignObject (setF, _) getF =
          (setF (); getF ())

    fun className (_, classNameRef) =
          !classNameRef
  end

(* This is what the clients needs to know about *)
signature CLASS =
  sig
    type t

    type object = ClassesImpl.object

    val create : t -> object
    val className : object -> string
    
    (* The method c forces a typecast and raises
       an exception if it fails *)
    val c : object -> t

    (* These 2 methods are for runtime determined forms of typecast *)
    val isA : object -> bool
    val tryTypecast : object -> t option
  end

(* This functor is applied to add a new type (class)
   to this "class system". *)
functor FClass(type newType
               val className : string) : CLASS =
  struct

    type t = newType
    type object = ClassesImpl.object

    val cell = ref(NONE : t option)
    val classNameRef = ref(className) (* FIXME: Use StringRes here *)

    fun setData value () =
          cell := SOME(value)

    fun tryGetData f () =
          case !cell of
            NONE => false
          | SOME(value) =>
              Util.finalize
                (fn _ => (f value; true))
                (fn _ => cell := NONE)

    fun create anObject =
          ClassesImpl.createObject
            (setData anObject)
            classNameRef

    fun isA anObject =
          ClassesImpl.tryAssignObject
            anObject (tryGetData (fn _ => ()))

    fun tryTypecast anObject =
          let
            val tmp = ref(NONE : t option)
            val _ = ClassesImpl.tryAssignObject anObject
                  (tryGetData (fn a => tmp := SOME(a)))
          in
            !tmp
          end

    fun className anObject =
          ClassesImpl.className anObject

    fun c anObject =
          case tryTypecast anObject of
            NONE => raise Excn.eRaiseGeneric
                      ("TYPECAST",
                       "Cannot typecast an object of type " ^
                       (className anObject) ^ " into type " ^
                       (!classNameRef))
          | SOME(value) => value
  end
